/*
   ESP32C3_WeatherStation_Rev2a.ino
   Rev2a add connection timeout
   ESP32C3 BLE_Scanner ServerTCP UDP_NTP
   by Matthew Ford,  2021/12/06
   (c)2021 Forward Computing and Control Pty. Ltd.
   NSW, Australia  www.forward.com.au
   This code may be freely used for both private and commerical use.
   Provide this copyright is maintained.

*/

// Needs Huge APP, 3M APP (no OTA) 1M SPIFF  Partition setting

#include <BLEDevice.h>
#include <BLEUtils.h>
#include <BLEScan.h>
#include <BLEAdvertisedDevice.h>

#include <WiFiClient.h>
#include <WebServer.h>
#include "ntpSupport.h"
#include "display.h"
#include <pfodDelay.h>


//#define DEBUG
static Stream *debugPtr = NULL;

// uncomment this to see all the scanning results
//#define DEBUG_BLE_SCANNING

WebServer server(80);
void startWebServer();
void start_pfodServer();
void handle_pfodServerConnection();

#include "ESPAutoWiFiConfig.h"  // download ESPAutoWiFiConfig.zip from  https://www.forward.com.au/pfod/ESPAutoWiFiConfig/index.html
#include "pfodLinkedPointerList.h" // iterable linked list of pointers to objects
#include "LastSeen.h"  // class for storing on linked list
#include "SafeString.h"
#include <WiFi.h>

// settings for the Auto WiFi connect
bool highForLedOn = true; // turns led OFF when WiFi connected
//and, -ve indicates the ws2812 driver is to be used
int ledPin = -8; // for ESP32-C3 ws2818 pin
size_t eepromOffset = 40; // if you use EEPROM.begin(size) in your code add the size here so AutoWiFi data is written after your data
// need to add ntpSupport getTimeZoneStorageSize() to this offset

SemaphoreHandle_t listOfLastSeen_Semaphore; // used to copy from listOf_BLE_LastSeen to listOfLastSeen
bool listUpdated = false; // set to true when copy from listOf_BLE_LastSeen to listOfLastSeen done, cleared when listOfLastSeen processed  // protected by listOfLastSeen_Semaphore

pfodDelay connectionTimeOutDelay;
const unsigned long CONNECTION_TIME_OUT_ms = 15000; // 15 sec,   pfodApp sends keepAlives every 5sec by default, set in pfodApp connection setup screen

// only BLE thread writes to these structs,
// loop() only reads from these and copies data for updates
// if BLE has lock on struct, loop() skips read and uses previous data
pfodLinkedPointerList<LastSeen> listOf_BLE_LastSeen; // updated by BLE
pfodLinkedPointerList<LastSeen> listOfLastSeen; // filled from BLE after each scan

// returns NULL if not found OR if list Ptr is NULL OR name is NULL
LastSeen* getLastSeen(const char* name, pfodLinkedPointerList<LastSeen>* listPtr) {
  if (!listPtr) {
    return NULL;
  }
  if (!name) {
    return NULL;
  }
  cSF(sfName, strlen(name));
  sfName = name;
  return getLastSeen(sfName, listPtr);
}

// returns NULL if not found OR if list Ptr is NULL
LastSeen* getLastSeen(SafeString &name, pfodLinkedPointerList<LastSeen>* list) {
  if (!list) {
    return NULL;
  }
  LastSeen *devicePtr = list->getFirst();
  while (devicePtr) {
    if (name.startsWith(devicePtr->getDeviceName())) {
      return devicePtr;
    } // else
    devicePtr = list->getNext();
  }
  return devicePtr; // null not found
}

int scanTime = 2; //In seconds
BLEScan *pBLEScan;
const unsigned long SKIP_TIME_MS = 11ul * 1000; // skip adverts for 11sec after first one
int devicesFoundThisScan = 0;
class MyAdvertisedDeviceCallbacks : public BLEAdvertisedDeviceCallbacks {
    void onResult(BLEAdvertisedDevice advertisedDevice) {
      if (!advertisedDevice.haveName()) {
        return;
      }
      // else
      cSF(sfName, 50); // max 32
      sfName = advertisedDevice.getName().c_str();
      sfName.trim();
      // search listOfLastSeen
      LastSeen *devicePtr = getLastSeen(sfName, &listOf_BLE_LastSeen);
      if (!devicePtr) {
        // not found
#ifdef DEBUG_BLE_SCANNING
        if (debugPtr) {
          debugPtr->print("Not scanning for "); debugPtr->println(sfName);
        }
#endif
        return;
      }
      // else found it check if seen recently
      unsigned long now = millis();
      if ((now - devicePtr->getLastSeen()) > SKIP_TIME_MS) {
        // need to update this
        devicePtr->updateLastSeen(now);
        devicePtr->setAdvertisedName(sfName.c_str()); // save the full advertised data
        if (debugPtr) {
          debugPtr->print("Scan found "); debugPtr->println(sfName);
        }
        devicesFoundThisScan++;
      } else {
#ifdef DEBUG_BLE_SCANNING
        if (debugPtr) {
          debugPtr->print("Skip repeat advertising for "); debugPtr->print(sfName); debugPtr->println(" seen recently");
        }
#endif
      }
    }
};

static TaskHandle_t bleScannerHandle = NULL;


void BLE_init() {
  BLEDevice::init("");
  pBLEScan = BLEDevice::getScan(); //create new scan
  pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
  pBLEScan->setActiveScan(true);
  pBLEScan->setInterval(40);  // same as Apple in foreground
  pBLEScan->setWindow(30); // same as Apple in foreground, less or equal setInterval value
}

// this call is protected by listOfLastSeen_Semaphore (if non-null)
void copyListOfLastSeen() { // sets listUpdated if updated
  // update listOfLastSeen for WeatherStation
  // use semaphore here, this is certainly not an ISR
  LastSeen *BLE_devicePtr = listOf_BLE_LastSeen.getFirst();
  while (BLE_devicePtr != NULL) {
    LastSeen *devicePtr = getLastSeen(BLE_devicePtr->getDeviceName(), &listOfLastSeen);
    if (!devicePtr) {
      // add it
      devicePtr = new LastSeen(BLE_devicePtr->getDeviceName());
      listOfLastSeen.add(devicePtr); // addit to WeatherStation list
    }
    // copy from BLE to listOfLastSeen if different
    if (devicePtr->getLastSeen() != BLE_devicePtr->getLastSeen()) {
      devicePtr->updateLastSeen(BLE_devicePtr->getLastSeen());
      devicePtr->setAdvertisedName(BLE_devicePtr->getAdvertisedName());
      listUpdated = true;
    }
    BLE_devicePtr = listOf_BLE_LastSeen.getNext();
  }
  devicesFoundThisScan = 0; // have copied the devices found, interested devices only
}


/* this function will be invoked when additionalTask was created */
void bleScannerTask( void * parameter ) {
  BLE_init();
  /* loop forever */
  devicesFoundThisScan = 0; // have copied the devices found, interested devices only
  for (;;) {
    // put your task code here, to run repeatedly:
    pBLEScan->clearResults(); // delete results fromBLEScan buffer to release memory
    BLEScanResults foundDevices = pBLEScan->start(scanTime, false);
    yield();
    if (devicesFoundThisScan) {
#ifdef DEBUG_BLE_SCANNING
      if (debugPtr) {
        debugPtr->print("Devices found: ");
        debugPtr->println(devicesFoundThisScan);
      }
#endif
      if (!listOfLastSeen_Semaphore) {
        // just copy
        copyListOfLastSeen(); // clears devicesFoundThisScan
      } else {// take it wait a little
        if ( xSemaphoreTake( listOfLastSeen_Semaphore, ( TickType_t ) 10 ) == pdTRUE ) {
          copyListOfLastSeen(); // clears devicesFoundThisScan
          xSemaphoreGive( listOfLastSeen_Semaphore );
        } // else devicesFoundThisScan not cleared will try and copy next time around
      }
    }
    delay(100); // use delay instead of yield() as yield() can give reboots
  }
  /* delete a task when finish,
    this will never happen because this is infinity loop */
  vTaskDelete( NULL );
}

void setUpWiFiServices() {
  setNtpSupportDebug(debugPtr);
  initializeNtpSupport();
  resetDefaultTZstr(); // only need this first time through
  startWebServer();
  start_pfodServer();
}
void setup() {
  Serial.begin(115200);

#ifdef DEBUG
  SafeString::setOutput(Serial); // for debug error msgs
  debugPtr = &Serial;
  debugPtr->println();
#endif
  /* Initialise wifi module */
#ifdef DEBUG
  setESPAutoWiFiConfigDebugOut(Serial); // turns on debug output for the ESPAutoWiFiConfig code
#endif

  if (ESPAutoWiFiConfigSetup(ledPin, highForLedOn, eepromOffset)) { // check if we should start access point to configure WiFi settings
    return; // in config mode so skip rest of setup
  }

  setDisplayDebug(debugPtr); // debugPtr can be NULL OK
  initializeDisplay(); // for pfodApp

  /* Create a mutex type semaphore. */
  listOfLastSeen_Semaphore = xSemaphoreCreateMutex();
  if (!listOfLastSeen_Semaphore) {
    if (debugPtr) {
      debugPtr->println("xSemaphoreCreateMutex failed, just copy without mutex protection");
    }
  }

  // add the names we are interested in NOTE the trailing , so that T_1, does not match T_11,
  LastSeen *devicePtr = new LastSeen(getOutsideDevice()); // note MUST use new since pfodLinkedPointerList uses delete when remove() called
  listOf_BLE_LastSeen.add(devicePtr);
  // add other devices here that you want to scan for
  devicePtr = new LastSeen(getInsideDevice());
  listOf_BLE_LastSeen.add(devicePtr);

  if (debugPtr) {
    debugPtr->println(" ================ ");
    debugPtr->print("Scanning for ");
    LastSeen *lastSeenPtr = listOf_BLE_LastSeen.getFirst();
    if (lastSeenPtr == NULL) {
      debugPtr->println(" Error: no devices specified.");
    } else {
      while (lastSeenPtr != NULL) {
        debugPtr->print(lastSeenPtr->getDeviceName()); debugPtr->print(" ");
        lastSeenPtr = listOf_BLE_LastSeen.getNext();
      }
    }
    debugPtr->println();
    debugPtr->println(" ================ ");
  }

  setUpWiFiServices(); // do this first. Get continual reboots if create BLE task first and then call this
  BaseType_t err = xTaskCreate(
                     bleScannerTask, /* Task function. */
                     "bleScannerTask", /* name of task. */
                     102400, /* Stack size of task */
                     NULL, /* parameter of the task */
                     1, /* priority of the task */ // the loop task also has priority 1 also so will share
                     &bleScannerHandle); /* Task handle to keep track of created task */
  if (((int)err) != 1) { // pPass ,  pdTRUE      ( ( BaseType_t ) 1 ) in #include "freertos/projdefs.h"
    if (debugPtr) {
      debugPtr->print("xTaskCreatePinnedToCore returned:");
      debugPtr->println((int)err);
    }
  }
}



void loop() {
  //  .. other code that MUST run all the time
  if (ESPAutoWiFiConfigLoop()) {  // handle WiFi config webpages
    return;  // skip the rest of the loop until config finished
  }
  if (!listOfLastSeen_Semaphore) {
    // just copy
    if (listUpdated) {
      listUpdated = false;
      processListOfLastSeen(listOfLastSeen); //
    }

  } else {// take it wait a little
    if ( xSemaphoreTake( listOfLastSeen_Semaphore, ( TickType_t ) 10 ) == pdTRUE ) {
      if (listUpdated) {
        listUpdated = false;
        processListOfLastSeen(listOfLastSeen);
      }
      xSemaphoreGive(listOfLastSeen_Semaphore );
    } // else devicesFoundThisScan not cleared will try and copy next time around
  }
  yield();
  server.handleClient();
  yield();
  processDisplay(); // needs to be called each loop
  yield();
  processNTP();
  yield();
  handle_pfodServerConnection();
  yield();
}

// ------------------ trival telnet server -------------------------
//how many clients should be able to telnet to this ESP32
#define MAX_SRV_CLIENTS 1

WiFiServer pfodServer(4989);
WiFiClient pfodServerClients[MAX_SRV_CLIENTS];

void start_pfodServer() {
  pfodServer.begin();
  pfodServer.setNoDelay(true);
  if (debugPtr) {
    debugPtr->print("pfodServer Ready to Use on ");
    debugPtr->print(WiFi.localIP());
    debugPtr->println(":4989");
  }
}

void handle_pfodServerConnection() {
  if (connectionTimeOutDelay.justFinished()) {
   if (debugPtr) {
      debugPtr->println("Connection timedOut, close");
   }
    for (int i = 0; i < MAX_SRV_CLIENTS; i++) {
      if (pfodServerClients[i]) {
        pfodServerClients[i].stop();
      }
    }
  }
  
  //check if there are any new clients
  uint8_t i;
  if (pfodServer.hasClient()) {
    for (i = 0; i < MAX_SRV_CLIENTS; i++) {
      //find free/disconnected spot
      if (!pfodServerClients[i] || !pfodServerClients[i].connected()) {
        if (pfodServerClients[i]) {
          pfodServerClients[i].stop();
        }
        pfodServerClients[i] = pfodServer.available();
        if (!pfodServerClients[i]) {
          if (debugPtr) {
            debugPtr->println("available broken");
          }
        } else {
          if (debugPtr) {
            debugPtr->print("New client: ");
            debugPtr->print(i); debugPtr->print(' ');
            debugPtr->println(pfodServerClients[i].remoteIP());
          }
          connectionTimeOutDelay.start(CONNECTION_TIME_OUT_ms);
          connect_pfodParser(&pfodServerClients[i]);
          break;
        }
      }
    }
    if (i >= MAX_SRV_CLIENTS) {
      //no free/disconnected spot so reject
      pfodServer.available().stop();
    }
  }
  //check clients for data
  for (i = 0; i < MAX_SRV_CLIENTS; i++) {
    if (pfodServerClients[i] && pfodServerClients[i].connected()) {
      handle_pfodParser();
    }
    else {
      if (pfodServerClients[i]) {
        pfodServerClients[i].stop();
      }
    }
  }
}


// ========== trivial webserver example =============
void notFound() {
  server.send(404, "text/plain", "Not found");
}

void handleRoot() {
  if (debugPtr) {
    debugPtr->println(">>> WebServer handleRoot");
  }


  String msg =
    "<html>\
  <head>\
    <meta http-equiv='refresh' content='5'/>\
    <title>BLE Temperature Sensors</title>\
    <style>\
      body { background-color: #cccccc; font-family: Arial, Helvetica, Sans-Serif; Color: #000088; }\
    </style>\
  </head>\
  <body>";
  time_t now;
  msg += "TimeZone: ";
  msg += getCurrentTZ();
  msg += "<br>";
  msg += getCurrentTZdescription();
  msg += "<p>At ";
  now = time(nullptr); // nullptr is a C++ null pointer literal which you can use anywhere you need to pass a null pointer.
  msg += ctime(&now);
  msg += "<br>";
  msg += "Time since reboot: ";
  msg += (millis()) / 1000.0;
  msg += " sec ago";
  msg += "<br>";
  msg += "The BLE devices found were:-<br>";
  // loop through and add reading found
  LastSeen *devicePtr = listOfLastSeen.getFirst();
  bool haveFoundDevice = false;
  msg += "<h1>";
  while (devicePtr != NULL) {
    const char* fullName = devicePtr->getAdvertisedName();
    if (*fullName) { // not empty fullName[0] != 0
      haveFoundDevice = true;
      msg += fullName;
      msg += "<font size=\"-1\"> ";
      msg += (millis() - devicePtr->getLastSeen()) / 1000.0;
      msg += " sec ago</font>";
      msg += "<br>";
    }
    devicePtr = listOfLastSeen.getNext();
  }
  if (!haveFoundDevice) {
    msg += "No devices found so far.<br>";
  }
  msg += "</h1>";
  msg += "</body></html>";

  server.send(200, "text/html", msg);
}

void startWebServer() {
  server.on("/", handleRoot);
  server.onNotFound(notFound);
  server.begin();
  if (debugPtr) {
    debugPtr->println("webserver started");
  }
}
