// Lux_Gauge.cpp
/*
   (c)2014-2022 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.
*/

#include "LuxGauge.h"
#include "SafeString.h"

// NOTE:  if sensor not in direct sun for more accurate temperature readins
// need to scale lux level see adjustmentFactor in setLux( ) method below
// idealy calibrate against a second sensor not in shadow

static const int sliderRange = 255; // scale to this length
static const float top_lux = 1800;
static const float bottom_lux = 0.55;
static const float top_log_lux = log10(top_lux);
static const float bottom_log_lux = log10(bottom_lux);
static const float log_lux_range = top_log_lux - bottom_log_lux;

LuxGauge::LuxGauge(pfodDwgs *_dwgsPtr) : pfodControl(_dwgsPtr) {
}

void LuxGauge::setLux(float _fvalue) {
  float adjustmentFactor = 2.0; // adjustment because sensor is in shadow
  fvalue = _fvalue * adjustmentFactor;

  float limited_fvalue = fvalue;
  if (limited_fvalue < bottom_lux) {
    limited_fvalue = bottom_lux;
  }
  if (limited_fvalue > top_lux) {
    limited_fvalue = top_lux;
  }
  logValue = log10(limited_fvalue);
}

float LuxGauge::logLuxToRange(float log_lx) {
  if (log_lx < bottom_log_lux) {
    log_lx = bottom_log_lux;
  }
  if (log_lx > top_log_lux) {
    log_lx = top_log_lux;
  }
  return (log_lx - bottom_log_lux) / log_lux_range * sliderRange;
}

void LuxGauge::drawScalePoint(float lx) {
  float log_lx = log10(lx);
  float pos = logLuxToRange(log_lx);
  cSF(sfText, 15);
  sfText.print(lx, 2, 4); // print float with max 2 dec but limited to 4 digits total
  dwgsPtr->label().fontSize(14).bold().text(sfText.c_str()).color(dwgsPtr->BLACK).offset(pos, 16).send();
  dwgsPtr->line().color(dwgsPtr->BLACK).offset(pos, 0).size(0, 7.5).send();
}
void LuxGauge::draw() {
  dwgsPtr->rectangle().filled().color(dwgsPtr->GREY).size(sliderRange, 1).offset(0, -0.5).send();
  drawScalePoint(1);
  drawScalePoint(10);
  drawScalePoint(100);
  drawScalePoint(1000);
  float pos = logLuxToRange(log10(1000));
  dwgsPtr->label().fontSize(12).text("Overcast").bold().color(dwgsPtr->BLACK).offset(pos, 30).send();
  pos = logLuxToRange(log10(100));
  dwgsPtr->label().fontSize(12).text("Dark Day").bold().color(dwgsPtr->BLACK).offset(pos, 30).send();
  pos = logLuxToRange(log10(10));
  dwgsPtr->label().fontSize(12).text("Sunset").bold().color(dwgsPtr->BLACK).offset(pos, 30).send();
  pos = logLuxToRange(log10(1));
  dwgsPtr->label().fontSize(12).text("Twilight").bold().color(dwgsPtr->BLACK).offset(pos, 30).send();
  dwgsPtr->index().idx(circle_idx).send();
  dwgsPtr->index().idx(bar_idx).send();
  dwgsPtr->index().idx(value_idx).send();
  update(); // update with current state
}

void LuxGauge::update() {
  if (fvalue < bottom_lux) {
    dwgsPtr->hide().idx(circle_idx).send();
    dwgsPtr->hide().idx(bar_idx).send();
  } else {
    dwgsPtr->unhide().idx(circle_idx).send();
    dwgsPtr->unhide().idx(bar_idx).send();
    dwgsPtr->circle().filled().radius(5).idx(circle_idx).color(220).offset(logLuxToRange(logValue), 0).send();
    dwgsPtr->rectangle().filled().idx(bar_idx).color(220).size(logLuxToRange(logValue), 4).offset(0, -2).send();
  }
  cSF(sfText, 15);
  if (fvalue < bottom_lux) {
    fvalue = bottom_lux;
    sfText = "<";
  }
  if (fvalue > top_lux) {
    fvalue = top_lux;
    sfText = ">";
  }
  sfText.print(fvalue, 2, 4); // print float with max 2 dec but limited to 4 digits total, right justified
  sfText += " lx";
  dwgsPtr->label().bold().text(sfText.c_str()).idx(value_idx).fontSize(24).color(dwgsPtr->RED).offset(128, -20).send();
}
