// TempGauge_F.cpp
/*
   (c)2022 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.
*/

#include "TempGauge_F.h"
#include "SafeString.h"


// _left true => scale to left current temp to right i.e. outside, false inside
TempGauge_F::TempGauge_F(pfodDwgs *_dwgsPtr, bool _left, int minTemp, int maxTemp, bool _addScaleLabels) : pfodControl(_dwgsPtr) {
  z_idx = pfodDwgs::reserveIdx(11); // z_idx to z_idx+8.   z_idx+1 and z_idx+2 not currently used
  sign = _left ? -1 : 1;
  bottom_temp = minTemp;
  top_temp = maxTemp;
  temp_range = top_temp - bottom_temp;
  addScaleLabels = _addScaleLabels;
}

void TempGauge_F::setDegF(float _fvalue) {
  if (_fvalue >= Temp24hrMaxMin::invalid) {
    // ignore
    return;
  }
  fvalue = _fvalue;
  maxMin24hr.setDeg(fvalue);
}

void TempGauge_F::setDegC(float _fvalue) {
  if (_fvalue >= Temp24hrMaxMin::invalid) {
    // ignore
    return;
  }
  // convert degC to degF for this gauge
  fvalue = _fvalue * 9 / 5 + 32;
  maxMin24hr.setDeg(fvalue);
}

// up from bottom so need to used the -ve of this return
float TempGauge_F::slidePosition(float value) {
  return (value - bottom_temp) * sliderRange / temp_range;
}
void TempGauge_F::drawScalePoint(int value) {
  if ((value < bottom_temp) || (value > top_temp)) {
    return;
  }
  cSF(sfLabel, 20);
  sfLabel = value;
  if (addScaleLabels) {
    dwgsPtr->label().fontSize(9).bold().text(sfLabel.c_str()).color(dwgsPtr->BLACK).offset(sign * 11, -slidePosition(value)).send();
  }
  dwgsPtr->line().color(dwgsPtr->BLACK).offset( sign * 4, -slidePosition(value)).size(-sign * 5, 0).send();
}

void TempGauge_F::drawScaleLabel(int value) {
  if ((value < bottom_temp) || (value > top_temp)) {
    return;
  }
  cSF(sfLabel, 20);
  sfLabel = value;
  dwgsPtr->label().fontSize(9).bold().text(sfLabel.c_str()).color(dwgsPtr->BLACK).offset(0, -slidePosition(value)).send();
}

static const int rangeChange = 70;
// NOTE: this method should match the drawScalePoint() calls below
void TempGauge_F::drawScale() {
  dwgsPtr->pushZero(0, 0, 0.36);  // scale down temp_range to 25 cols
  dwgsPtr->label().fontSize(12).bold().text("\342\204\211").color(dwgsPtr->BLACK).offset(0, -sliderRange - 10).send();
  if (temp_range > rangeChange) { // use 15 deg steps
    int tmpF = -45;
    while (tmpF <= 120) {
     drawScaleLabel(tmpF);
     tmpF += 15;       
    }
  } else {
    int tmpF = -40;
    while (tmpF <= 120) {
     drawScaleLabel(tmpF);
     tmpF += 10;       
    }
  }
  dwgsPtr->popZero();
}

void TempGauge_F::draw() {
  dwgsPtr->pushZero(0, 0, 0.36);  // scale down temp_range to 25 cols
  dwgsPtr->rectangle().filled().color(dwgsPtr->SILVER).size(2, -sliderRange).offset(-1, 0).send();
  if (temp_range > rangeChange) { // use 15 deg steps
    int tmpF = -45;
    while (tmpF <= 120) {
     drawScalePoint(tmpF);
     tmpF += 15;       
    }
  } else {
    int tmpF = -40;
    while (tmpF <= 120) {
     drawScalePoint(tmpF);
     tmpF += 10;       
    }
  }

  // defined idxs for later use
  dwgsPtr->index().idx(z_idx).send();
  dwgsPtr->index().idx(z_idx + 3).send();
  dwgsPtr->index().idx(z_idx + 4).send();
  dwgsPtr->index().idx(z_idx + 5).send();
  update(); // update with current state
  dwgsPtr->popZero();
}

float TempGauge_F::limitedFillSize(float fvalue) {
  float fillSize = slidePosition(fvalue);
  if (fillSize > sliderRange) {
    fillSize = sliderRange;
  }
  if (fillSize < 0) {
    fillSize = 0;
  }
  return fillSize;
}

void TempGauge_F::update() {
  float fillSize = limitedFillSize(fvalue);
  if (sign < 0) { // scale to left
    dwgsPtr->label().decimals(1).idx(z_idx + 8).left().bold().floatReading(fvalue).units(F("\342\204\211")).fontSize(13).color(dwgsPtr->RED).offset(-sign * 15, -fillSize).send();
    dwgsPtr->arc().filled().idx(z_idx + 5).color(dwgsPtr->RED).offset(-sign * 3, -fillSize).radius(10).start(-15).angle(30).send();
  } else {
    dwgsPtr->label().decimals(1).idx(z_idx + 8).right().bold().floatReading(fvalue).units(F("\342\204\211")).fontSize(13).color(dwgsPtr->RED).offset(-sign * 15, -fillSize).send();
    dwgsPtr->arc().filled().idx(z_idx + 5).color(dwgsPtr->RED).offset(-sign * 3, -fillSize).radius(10).start(195).angle(-30).send();
  }
  float max = maxMin24hr.getMax();// + 10;
  float min = maxMin24hr.getMin();// - 10;
  if ((max < 1000) && (min < 1000)) {
    float maxfillSize = limitedFillSize(max);
    float minfillSize = limitedFillSize(min);
    dwgsPtr->rectangle().filled().idx(z_idx).color(dwgsPtr->RED).size(4, -(maxfillSize - minfillSize)).offset(-2, -minfillSize).send();
    cSF(sfText_max, 20);
    sfText_max = "24hr max ";
    sfText_max.print(max, 1);
    sfText_max += "\342\204\211";
    cSF(sfText_min, 20);
    sfText_min = "24hr min ";
    sfText_min.print(min, 1);
    sfText_min += "\342\204\211";
    if (sign < 0) { // left
      if (fillSize > (sliderRange - 8)) {
        dwgsPtr->label().text(sfText_max.c_str()).decimals(1).idx(z_idx + 3).left().bold().fontSize(4).color(dwgsPtr->BLACK).offset(-sign * 6, -fillSize + 8).send();
        dwgsPtr->label().text(sfText_min.c_str()).decimals(1).idx(z_idx + 4).left().bold().fontSize(4).color(dwgsPtr->BLACK).offset(-sign * 6, 0).send();
      } else if (fillSize < 8) {
        dwgsPtr->label().text(sfText_max.c_str()).decimals(1).idx(z_idx + 3).left().bold().fontSize(4).color(dwgsPtr->BLACK).offset(-sign * 6, -sliderRange).send();
        dwgsPtr->label().text(sfText_min.c_str()).decimals(1).idx(z_idx + 4).left().bold().fontSize(4).color(dwgsPtr->BLACK).offset(-sign * 6, -fillSize - 8).send();
      } else {
        dwgsPtr->label().text(sfText_max.c_str()).decimals(1).idx(z_idx + 3).left().bold().fontSize(4).color(dwgsPtr->BLACK).offset(-sign * 6, -sliderRange).send();
        dwgsPtr->label().text(sfText_min.c_str()).decimals(1).idx(z_idx + 4).left().bold().fontSize(4).color(dwgsPtr->BLACK).offset(-sign * 6, 0).send();
      }
    } else {
      if (fillSize > (sliderRange - 8)) {
        dwgsPtr->label().text(sfText_max.c_str()).decimals(1).idx(z_idx + 3).right().bold().fontSize(4).color(dwgsPtr->BLACK).offset(-sign * 6, -fillSize + 8).send();
        dwgsPtr->label().text(sfText_min.c_str()).decimals(1).idx(z_idx + 4).right().bold().fontSize(4).color(dwgsPtr->BLACK).offset(-sign * 6, 0).send();
      } else if (fillSize < 8) {
        dwgsPtr->label().text(sfText_max.c_str()).decimals(1).idx(z_idx + 3).right().bold().fontSize(4).color(dwgsPtr->BLACK).offset(-sign * 6, -sliderRange).send();
        dwgsPtr->label().text(sfText_min.c_str()).decimals(1).idx(z_idx + 4).right().bold().fontSize(4).color(dwgsPtr->BLACK).offset(-sign * 6, -fillSize - 8).send();
      } else {
        dwgsPtr->label().text(sfText_max.c_str()).decimals(1).idx(z_idx + 3).right().bold().fontSize(4).color(dwgsPtr->BLACK).offset(-sign * 6, -sliderRange).send();
        dwgsPtr->label().text(sfText_min.c_str()).decimals(1).idx(z_idx + 4).right().bold().fontSize(4).color(dwgsPtr->BLACK).offset(-sign * 6, 0).send();
      }
    }
  }
}
