
/*
  display.cpp for WeatherStation
   (c)2022 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.
*/
// download the libraries from http://www.forward.com.au/pfod/pfodParserLibraries/index.html

/**
   This sketch accepts serial inputs on Serial consistion of BLE Senor names and values
   Looks for OUTSIDE_DEVICE and INSIDE_DEVICE sensors.
   OUTSIDE_DEVICE, WL_1 is the outside sensor, INSIDE_DEVICE, W_1 is the inside sensor
   WL_1 has a lux level, for a future project, in addition to temp,rh,pressure

  Sensor advertised names start with
  T_.. for Temp(C) sensors, H_.. for Temp(C),Humditiy(%) sensors, W_.. for Temp(C),Humdity(%),Barometer(hPa) sensors, WL_.. for Temp(C),Humdity(%),Barometer(hPa),Lux level(Lux)
  keep this LOCAL_NAME short as it uses up available advertising data space (see below)
  Format of advertized name
  For T_.. devices   LOCAL_NAME,degC
  For H_.. devices   LOCAL_NAME,degC,%RH
  For W_.. devices   LOCAL_NAME,degC,%RH,hPa
  For WL_.. devices  LOCAL_NAME,degC,%RH,hPa,Lux
*/
#include "Temp24hrMaxMin.h"

// download SafeString V4.1.5+ library from the Arduino Library manager or from
// https://www.forward.com.au/pfod/ArduinoProgramming/SafeString/index.html
#include "display.h"
#include "SafeString.h"
#include "millisDelay.h"
#include "ntpSupport.h" // for local time AM/PM
#include <WiFiClient.h>

#include "rh_Gauge.h"
#include "TempGauge_C.h"
// or
#include "TempGauge_F.h"

#include "LuxGauge.h"

#include "hPa_BarGraph.h"
#include "ChartButton.h"

#include <ESPBufferedClient.h>
#include <pfodParser.h>

extern pfodDelay connectionTimeOutDelay; // reset here on each msg

pfodParser parser("V1"); // create a parser with menu version string to handle the pfod messages
ESPBufferedClient bufferedClient;

pfodDwgs dwgs(&parser);  // drawing support

//hPa adjustment for altitude difference
float hPaOffset = +7.3; // from Bureau of Meterology, Aus. local reading 1013.0hPa (reduced to sea level) while WL_1 reading is 1005.7 => add 7.3 for reading adjusted for local elevation

// uncomment this line to get scale and plots in degs F
// otherwise default is degsC  in any case all readings from the BLE sensors are in degsC and the conversions to degsF done here
//#define DEGS_F 

#ifdef DEGS_F
// for degF gauge  NOTE: both scale must be set the same
TempGauge_F inside_temp(&dwgs, false,30,100,false); // scale points on right, current temp on left 30 to 100degF range, no labels, need to call drawScale() for labels
TempGauge_F outside_temp(&dwgs, true,30,100,false); // scale points on left, current temp on right
// to set the scale use TempGauge_C outside_temp(&dwgs, true/false, minTemp,maxTemp);
#else // DEGS C
// for degC gauge  NOTE: both scale must be set the same
TempGauge_C inside_temp(&dwgs, false, 0, 35, false); // scale points on right, current temp on left  0 to 35degC range, no labels, need to call drawScale() for labels
TempGauge_C outside_temp(&dwgs, true, 0, 35, false); // scale points on left, current temp on right  0 to 35degC range, no labels
// to set the scale use TempGauge_C outside_temp(&dwgs, true/false, minTemp,maxTemp);
#endif

// usually leave this commented out unless you have an outside lux reading
//#define LUX_GAUGE


//#define DEBUG
static Stream *debugPtr = NULL;

void setDisplayDebug(Stream* debugOutPtr) {
  debugPtr = debugOutPtr;
}

// These are the names of the BLE devices to listen to
static const char OUTSIDE_DEVICE[] = "WL_1,"; // note the trailing , so T_1, does not match T_11,
static const char INSIDE_DEVICE[] = "W_1,";

const char* getOutsideDevice(){
  return OUTSIDE_DEVICE; // name of outside BLE
}
const char* getInsideDevice() {
  return INSIDE_DEVICE; // name of inside BLE
}


cSF(sensorData, 80); // max name size is 32 so 80 is enough, allows for some debug from other side as well
const size_t MAX_BLE_NAME_SIZE = 32;
char T_DEVICE[] = "T_"; // T_ devices have temp only (in degC)
char H_DEVICE[] = "H_"; // H_ devices have temp(degC),RH(%)
char W_DEVICE[] = "W_"; // W_ devices have temp(degC),RH(%),AirPressure(hPa)
char WL_DEVICE[] = "WL_"; // WL_ devices have temp(degC),RH(%),AirPressure(hPa),Lux
cSF(sfDevice, MAX_BLE_NAME_SIZE);
cSF(sf_TempC, MAX_BLE_NAME_SIZE);
cSF(sf_RH, MAX_BLE_NAME_SIZE);
cSF(sf_hPa, MAX_BLE_NAME_SIZE);
cSF(sf_lx, MAX_BLE_NAME_SIZE);
float outside_tempC = Temp24hrMaxMin::invalid;
float RH_out = 0;
float hPa_out = 0;
float lx_out = 0.01;  // will always be this if not lux reading decoded
unsigned long lastSeen_out = 0;
float inside_tempC = Temp24hrMaxMin::invalid;
float RH_in = 0;
float hPa_in = 0; // not used
unsigned long lastSeen_in = 0;

// decode globals
float lx = 0.01; // from WL_1,
float tempC = 0; // from decode set to inside/outside later
float RH = 0; // from decode set to inside/outside later
float hPa = 0; // from decode set to inside/outside later

struct DataPoint {
  unsigned long t;
  int hPaTimes10;
  int inCTimes10;
  int outCTimes10;
  int inRH;
  int outRH;
};
static const size_t DAY_SUB_INTERVALS = 6; //
static const size_t ARRAY_SIZE_DAY = hPa_BarGraph::GRAPH_X_AXIS_SIZE * DAY_SUB_INTERVALS; // 5min sub-intervals
static const size_t ARRAY_SIZE = (3 * ARRAY_SIZE_DAY) / 2; // 36hrs
static size_t nextArrayIdx = 0;
const unsigned long DATA_TIMER_MS = hPa_BarGraph::X_AXIS_MS / ARRAY_SIZE_DAY;
static millisDelay dataTimer;
struct DataPoint data[ARRAY_SIZE];

void sendMainMenu();
void sendMainMenuUpdate();
void sendDrawingUpdates_y();
void sendDrawing_y(long l);
void sendDrawing_y0();
void sendDrawingUpdates_x();
void sendDrawing_x(long l);
void sendDrawing_x0();
void sendDrawing_x1();
void sendDrawingUpdates_z();
void sendDrawing_z(long l);
void sendDrawing_z0();
void sendDrawing_z1();
void sendDrawing_z2();
void sendDrawing_z3();
void sendDrawing_z4();
void sendDrawingUpdates_lx();
void sendDrawing_lx(long l);
void sendDrawing_lx0();
void closeConnection(Stream *io);
void updateGauges();

const int portNo = 4989; // What TCP port to listen on for connections.
// handled by ESPAutoWiFiConfig

WiFiClient client;
WiFiClient* clientPtr; // non-null if have connected client

unsigned long plot_msOffset = 0; // set by {@} response
bool clearPlot = false; // set by the {@} response code
bool sending_plotData = false;
void start_DataSend();
void send_Data();

const char chart_cmd = 'C'; // opens plots with touch

rh_Gauge in_rh(&dwgs);
rh_Gauge out_rh(&dwgs);
LuxGauge out_lx(&dwgs);
hPa_BarGraph hPa_graph(&dwgs);
ChartButton chartButton(&dwgs,chart_cmd); 


static bool displayInitialized = false;
static int z_idx; // for title updates
static void initializeDataPoints();
size_t incrementArrayIdx(size_t i);
size_t decrementArrayIdx(size_t i);

static void refresh_hPa_graphData(size_t lastDataAddedIdx) {
  // refresh hPa_graph data array with this latest data
 // Serial.println(); Serial.print(getCurrentTime_hhmmAM_PM());  Serial.println(" refresh_hPa_graphData");
  size_t arrayIdx = lastDataAddedIdx;
  for (size_t i = 0; i < hPa_graph.getArraySize(); i++) {
    hPa_graph.setArrayData(data[arrayIdx].t, data[arrayIdx].hPaTimes10, i);
    for (size_t k = 0; k < DAY_SUB_INTERVALS; k++) {
      arrayIdx = decrementArrayIdx(arrayIdx); // skip sub intervals
    }
  }
}

// needs to be called each loop
void processDisplay() {
  if (!displayInitialized) {
    return;
  }
  if (dataTimer.justFinished()) {
    dataTimer.restart();
    // save next data point
    DataPoint point;
    point.t = millis();
    point.hPaTimes10 = hPa_out * 10.0;
    point.inCTimes10 = inside_tempC * 10.0;
    point.outCTimes10 = outside_tempC * 10.0;
    point.inRH = RH_in;
    point.outRH = RH_out;
    data[nextArrayIdx] = point;
    yield();
    refresh_hPa_graphData(nextArrayIdx); // idx where data just stored
    nextArrayIdx = incrementArrayIdx(nextArrayIdx); // update to next one
  }
}

void initializeDisplay() {
  if (displayInitialized) {
    return;
  }
  initializeDataPoints();
  dataTimer.start(DATA_TIMER_MS);
  updateGauges();
  // get unique indices so the W_1 and WL_1 text can be identified for replacement/updating
  // see sendDrawingUpdates_z() below
  z_idx = pfodDwgs::reserveIdx(5); // z_idx to z_idx+4 for W_1, WL_1 title updates
  inside_temp.setLabel(F("Inside"));
  outside_temp.setLabel(F("Outside"));
  in_rh.setLabel(F("Inside"));
  out_rh.setLabel(F("Outside"));
  displayInitialized = true;
  outside_tempC = Temp24hrMaxMin::invalid;
  RH_out = 0;
  hPa_out = 0;
  inside_tempC = Temp24hrMaxMin::invalid;
  RH_in = 0;
  hPa_in = 0;
}

static void initializeDataPoints() {
  DataPoint point;
  point.hPaTimes10 = hPa_BarGraph::invalid;
  point.inCTimes10 = Temp24hrMaxMin::invalid * 10;
  point.outCTimes10 = Temp24hrMaxMin::invalid * 10;
  point.inRH = 0;
  point.outRH = 0;
  unsigned long ms = millis();
  for (size_t i = 0; i < ARRAY_SIZE; i++) {
    point.t = ms - i * DATA_TIMER_MS;
    data[i] = point;
  }
}

size_t incrementArrayIdx(size_t i) {
  i++;
  if (i >= ARRAY_SIZE) {
    i = 0;
  }
  return i;
}

size_t decrementArrayIdx(size_t i) {
  if (i == 0) {
    i = ARRAY_SIZE;
  }
  i--;
  return i;
}

// Commands for drawing controls only need to unique per drawing, you can re-use cmd 'a' in another drawing.

// appends to sfResult if >1 min since last update
void getLastSeenMins(SafeString &sfResult, unsigned long ms) {
  if (ms == 0) {
    return; // not seen yet
  }
  unsigned long delta_ms = millis() - ms;
  int mins = (int)(delta_ms / (60 * 1000)); // minutes
  if (mins > 1) { // only show if miss one
    sfResult += "\n";
    sfResult += mins;
    sfResult += " mins ago";
  }
  return;
}

bool checkValidName(SafeString &sfName) {
  return (sfName.startsWith(W_DEVICE) || sfName.startsWith(WL_DEVICE)); // omit sfName.startsWith(T_DEVICE) || sfName.startsWith(H_DEVICE) ||
}

bool decode(SafeString &sfNameIn);

// this call is protected by listOfLastSeen_Semaphore (if non-null)
void processListOfLastSeen(pfodLinkedPointerList<LastSeen> &lastSeenList) {
  // update listOfLastSeen for WeatherStation
  // use semaphore here, this is certainly not an ISR
  LastSeen *devicePtr = lastSeenList.getFirst();
  while (devicePtr != NULL) {
    if (devicePtr->wasUpdated()) { // this call clears device updated flag
      // process these new values
      if (debugPtr) {
        debugPtr->print("Process: "); debugPtr->println(devicePtr->getAdvertisedName());
      }
      cSF(sensorData, strlen(devicePtr->getAdvertisedName()));
      sensorData = devicePtr->getAdvertisedName();
      if (checkValidName(sensorData)) { // is it W_1, or WL_1,
        if (decode(sensorData)) { // decode consumes sensorData string
          if (sensorData.startsWith(getOutsideDevice())) { // if WL_1 then outside data
            outside_tempC = tempC;
            RH_out = RH;
            hPa_out = hPa + hPaOffset;
            lx_out = lx * 2; // double lx to allow for mounting in shadow
            // lx will be just 0.01 if no reading decoded.
            lastSeen_out = devicePtr->getLastSeen();
          } else { // else inside data
            inside_tempC = tempC;
            RH_in = RH;
            hPa_in = hPa + hPaOffset;
            lastSeen_in = devicePtr->getLastSeen();
          }
          updateGauges();
          // uses the gobals updated if returns true
        }
      }
    }
    devicePtr = lastSeenList.getNext();
  }
}

bool decode(const char* nameIn) {
  cSF(sfNameIn, strlen(nameIn));
  sfNameIn = nameIn;
  return decode(sfNameIn);
}

bool decode(SafeString &sfNameIn) {
  sfNameIn.trim();
  if (!checkValidName(sfNameIn)) {
    return false;
  }
  cSF(sfName, sfNameIn.length());
  sfName = sfNameIn;
  // no need to clear the previous tokens as nextToken does that
  if (sfName.startsWith(T_DEVICE)) {
    // skip first field and pick up temp
    sfName.nextToken(sfDevice, ','); // remove name
    sfName.nextToken(sf_TempC, ','); // pick up temp
    if (debugPtr) {
      debugPtr->print(sfDevice); debugPtr->print(" Temperature:"); debugPtr->print(sf_TempC); debugPtr->println("C");
    }
    // convert Temp
    if (!sf_TempC.toFloat(tempC)) {
      return false; // last tempC unchanged
    }
    inside_tempC = tempC; // set global
  } else if (sfName.startsWith(H_DEVICE)) {
    // skip first field and pick up temp
    sfName.nextToken(sfDevice, ','); // remove name
    sfName.nextToken(sf_TempC, ','); // pickup Temp
    sfName.nextToken(sf_RH, ','); // pickup RH
    if (debugPtr) {
      debugPtr->print(sfDevice);
      debugPtr->print(" Temperature:"); debugPtr->print(sf_TempC); debugPtr->print("C");
      debugPtr->print(" RH:"); debugPtr->print(sf_RH); debugPtr->print("%");
      debugPtr->println();
    }
    if (!sf_TempC.toFloat(tempC)) {
      return false; // not a valid float, last tempC unchanged
    }
    if (!sf_RH.toFloat(RH)) {
      return false; // not a valid float, last RH unchanged
    }
  } else if (sfName.startsWith(W_DEVICE)) {
    // skip first field and pick up temp
    sfName.nextToken(sfDevice, ','); // remove name
    sfName.nextToken(sf_TempC, ','); // pickup Temp
    sfName.nextToken(sf_RH, ','); // pickup RH
    sfName.nextToken(sf_hPa, ','); // pickup Air Pressure
    if (debugPtr) {
      debugPtr->print(sfDevice);
      debugPtr->print(" Temperature:"); debugPtr->print(sf_TempC); debugPtr->print("C");
      debugPtr->print(" RH:"); debugPtr->print(sf_RH); debugPtr->print("%");
      debugPtr->print(" Barometer:"); debugPtr->print(sf_hPa); debugPtr->print(" hPa");
      debugPtr->println();
    }
    if (!sf_TempC.toFloat(tempC)) {
      return false; // not a valid float, last tempC unchanged
    }
    if (!sf_RH.toFloat(RH)) {
      return false; // not a valid float, last RH unchanged
    }
    if (!sf_hPa.toFloat(hPa)) {
      return false; // not a valid float, last hPA unchanged
    }
  } else if (sfName.startsWith(WL_DEVICE)) {
    // skip first field and pick up temp
    sfName.nextToken(sfDevice, ','); // remove name
    sfName.nextToken(sf_TempC, ','); // pickup Temp
    sfName.nextToken(sf_RH, ','); // pickup RH
    sfName.nextToken(sf_hPa, ','); // pickup Air Pressure
    sfName.nextToken(sf_lx, ','); // pickup Air Pressure
    if (debugPtr) {
      debugPtr->print(sfDevice);
      debugPtr->print(" Temperature:"); debugPtr->print(sf_TempC); debugPtr->print("C");
      debugPtr->print(" RH:"); debugPtr->print(sf_RH); debugPtr->print("%");
      debugPtr->print(" Barometer:"); debugPtr->print(sf_hPa); debugPtr->print(" hPa");
      debugPtr->print(" Lux:"); debugPtr->print(sf_lx); debugPtr->print(" lx");
      debugPtr->println();
    }
    if (!sf_TempC.toFloat(tempC)) {
      return false; // not a valid float,  last tempC unchanged
    }
    if (!sf_RH.toFloat(RH)) {
      return false; // not a valid float,  last RH unchanged
    }
    if (!sf_hPa.toFloat(hPa)) {
      return false; // not a valid float,  last hPA unchanged
    }
    if (!sf_lx.toFloat(lx)) {
      return false; // not a valid float,  last lx unchanged
    }
  } else {
    if (debugPtr) {
      debugPtr->print("Error: Device not supported -- "); debugPtr->println(sfName);
    }
    return false;
  }
  return true;
}

void updateGauges() {
  out_lx.setLux(lx_out);
  in_rh.setValue(RH_in);
  out_rh.setValue(RH_out);
  outside_temp.setDegC(outside_tempC);  // if temp from sensor is in degF use setDegF instead
  inside_temp.setDegC(inside_tempC);   // if temp from sensor is in degF use setDegF instead
  hPa_graph.set_hPa(hPa_out);
  if (debugPtr) {
    debugPtr->print("updateGauges: ");
    debugPtr->print(" Outside Temp:"); debugPtr->print(outside_tempC); debugPtr->print("C");
    debugPtr->print(" RH:"); debugPtr->print(RH_out); debugPtr->print("%");
    debugPtr->print(" Barometer:"); debugPtr->print(hPa_out); debugPtr->print(" hPa");
    debugPtr->print(" Lux:"); debugPtr->print(lx_out); debugPtr->print(" lx");
    debugPtr->println();
    debugPtr->print(" Inside Temp:"); debugPtr->print(inside_tempC); debugPtr->print("C");
    debugPtr->print(" RH:"); debugPtr->print(RH_in); debugPtr->print("%");
    debugPtr->print(" Barometer:"); debugPtr->print(hPa_in); debugPtr->print(" hPa");
    debugPtr->println();
  }
}


void connect_pfodParser(WiFiClient * client) {
  parser.connect(bufferedClient.connect(client)); // sets new io stream to read from and write to
}

void handle_pfodParser() {
  if (!displayInitialized) {
    if (debugPtr) {
      debugPtr->println("Error: Call initializeDisplay() first.");
    }
    return;
  }
  uint8_t cmd = parser.parse(); // parse incoming data from connection
  // parser returns non-zero when a pfod command is fully parsed
  if (cmd != 0) { // have parsed a complete msg { to }
    connectionTimeOutDelay.restart();  // restart timeout on each new msg,  pfodApp sends keepAlives every 5sec by default
    uint8_t* pfodFirstArg = parser.getFirstArg(); // may point to \0 if no arguments in this msg.
    pfod_MAYBE_UNUSED(pfodFirstArg); // may not be used, just suppress warning
    long pfodLongRtn; // used for parsing long return arguments, if any
    pfod_MAYBE_UNUSED(pfodLongRtn); // may not be used, just suppress warning
    if ('.' == cmd) {
      sending_plotData = false; //stop sending data
      // pfodApp has connected and sent {.} , it is asking for the main menu
      if (!parser.isRefresh()) {
        sendMainMenu(); // send back the menu designed
      } else {
        sendMainMenuUpdate(); // menu is cached just send update
      }

      // handle {@} request
    } else if ('@' == cmd) { // pfodApp requested 'current' time
      plot_msOffset = millis(); // capture current millis as offset rawdata timestamps
      clearPlot = true; // clear plot on reconnect as have new plot_msOffset
      parser.print(F("{@`0}")); // return `0 as 'current' raw data milliseconds

      // now handle commands returned from button/sliders
    } else if ('B' == cmd) { // user pressed menu item that loaded drawing with load cmd 'lx'
      // in the main Menu of Menu_1
      sendDrawingUpdates_lx(); // update the drawing

    } else if ('A' == cmd) { // user pressed menu item that loaded drawing with load cmd 'z'
      // in the main Menu of Menu_1
      byte dwgCmd = parser.parseDwgCmd();  // parse rest of dwgCmd, return first char of active cmd
      if ((dwgCmd == chart_cmd)) { // the click / touch filtering happens in the pfodApp to filter when to send these cmds
        //Serial.println("Chart_cmd");
        // open plot window and send latest data
        parser.print(F("{=In/Out Temp/RH and Barometric Pressure`"));
        parser.print(ARRAY_SIZE); // plot all the data, default is to only plot last 500
        parser.print(F("~E HH:mm"));
        if (!sending_plotData) {
          parser.print(F("~C")); // clear last plot data
        }
        #ifdef DEGS_F
        parser.print(F("|date`0|hPa~1030~950~hPa`3|Temp In~100~30~\342\204\211`1|Temp Out~100~30~\342\204\211`1|RH In~100~0~RH%`2|RH Out~100~0~RH%`2}"));
        #else
        parser.print(F("|date`0|hPa~1030~950~hPa`3|Temp In~35~0~\342\204\203`1|Temp Out~35~0~\342\204\203`1|RH In~100~0~RH%`2|RH Out~100~0~RH%`2}"));
        #endif
        // three plots against day/time, first(`1) one inside/outside temp, second(`2) is RH in/out and third(`3) plots hPa
        // plots autmatically expand the yAxis if the values exceed the plot range.
        start_DataSend();
      } else { //  ignore and update
        sendDrawingUpdates_z(); // update the drawing
      }

    } else if ('y' == cmd) { // pfodApp is asking to load dwg 'y'
      if (!parser.isRefresh()) { // not refresh, send whole dwg
        parser.parseLong(pfodFirstArg, &pfodLongRtn); // parse first arg as a long
        sendDrawing_y(pfodLongRtn);
      } else { // refresh just update drawing state
        sendDrawingUpdates_y();
      }

    } else if (parser.cmdEquals("lx")) { // pfodApp is asking to load dwg "lx"
      if (!parser.isRefresh()) { // not refresh, send whole dwg
        parser.parseLong(pfodFirstArg, &pfodLongRtn); // parse first arg as a long
        sendDrawing_lx(pfodLongRtn);
      } else { // refresh just update drawing state
        sendDrawingUpdates_lx();
      }

    } else if ('x' == cmd) { // pfodApp is asking to load dwg 'x'
      if (!parser.isRefresh()) { // not refresh, send whole dwg
        parser.parseLong(pfodFirstArg, &pfodLongRtn); // parse first arg as a long
        sendDrawing_x(pfodLongRtn);
      } else { // refresh just update drawing state
        sendDrawingUpdates_x();
      }

    } else if ('z' == cmd) { // pfodApp is asking to load dwg 'z'
      if (!parser.isRefresh()) { // not refresh, send whole dwg
        parser.parseLong(pfodFirstArg, &pfodLongRtn); // parse first arg as a long
        sendDrawing_z(pfodLongRtn);
      } else { // refresh just update drawing state
        sendDrawingUpdates_z();
      }

    } else if ('!' == cmd) {
      // CloseConnection command
      closeConnection(parser.getPfodAppStream());
    } else {
      // unknown command
      parser.print(F("{}")); // always send back a pfod msg otherwise pfodApp will disconnect.
    }
  }
  send_Data();
}

size_t printIdx, nextIdx;
cSF(sfData, 80); // large enough for millis(),hPa,inTemp,outTemp\r\n

// inT and outT here are in degsC
void fillDataOut(SafeString &sfData, unsigned long t, float hPa, float inT, float outT, int inRH, int outRH) {
  sfData.clear();
  sfData.print((int32_t)(t  - plot_msOffset)); // time in mS +/-, previous times are -ve
  sfData.print(',');
  if (hPa <= hPa_BarGraph::invalid) {
    sfData.print("---");
  } else {
    sfData.print(hPa);
  }
  sfData.print(',');
  if (inT >= Temp24hrMaxMin::invalid) {
    sfData.print("---");
  } else {
    #ifdef DEGS_F
    sfData.print(inT*9/5 + 32); // in degsF
    #else
    sfData.print(inT);
    #endif
  }
  sfData.print(',');
  if (outT >= Temp24hrMaxMin::invalid) {
    sfData.print("---");
  } else {
    #ifdef DEGS_F
    sfData.print(outT*9/5 + 32);  // in degsF
    #else
    sfData.print(outT);
    #endif
  }
  sfData.print(',');
  if (inRH <=0) { // invalid
    sfData.print("---");
  } else {
    sfData.print(inRH);
  }
  sfData.print(',');
  if (outRH <=0) { // invalid
    sfData.print("---");
  } else {
    sfData.print(outRH);
  }
  sfData.println(); // end of CSV data record
}

void start_DataSend() {
  if (sending_plotData) {
    return; // already sending
  }
  sending_plotData = true;
  nextIdx = nextArrayIdx;
  // go back one
  printIdx = decrementArrayIdx(nextIdx);
  // send current value
  fillDataOut(sfData, millis(), hPa_out, inside_tempC, outside_tempC, RH_in, RH_out);
  if (hPa_out < hPa_BarGraph::invalid) {
    // skip
  } else {
    parser.print(sfData);  //
   // Serial.println(sfData);
  }
  send_Data();
}


// sends one block of data at a time
void send_Data() {
  if (!sending_plotData) {
    return;
  }
  while (printIdx != nextIdx) {
    fillDataOut(sfData, data[printIdx].t, ((float)data[printIdx].hPaTimes10) / 10.0, ((float)data[printIdx].inCTimes10) / 10.0, ((float)data[printIdx].outCTimes10) / 10.0, 
    data[printIdx].inRH, data[printIdx].outRH);
    if ((data[printIdx].inRH <=0) && (data[printIdx].outRH <= 0)) { // no data at all yet
      // skip
    } else {
      if (bufferedClient.availableForWrite() < ((int)sfData.length())) { // availableForWrite is an int !!*@#
        parser.flush(); // blocks here until WiFi send this block of data.
        break; // do not increment printIdx, send it next time
      }
      parser.print(sfData);  //
   //   Serial.println(sfData);
    }
    printIdx = decrementArrayIdx(printIdx);
  }
  if (printIdx == nextIdx) {
    sending_plotData = false; // can start again
  }
}

void closeConnection(Stream *io) {
  (void)(io); // unused
  // add any special code here to force connection to be dropped
  parser.closeConnection(); // nulls io stream
  bufferedClient.stop(); // clears client reference
  //  clientPtr->stop();
  //  clientPtr = NULL;
}

void sendMainMenu() {
  // !! Remember to change the parser version string
  //    every time you edit this method
  parser.print(F("{,"));  // start a Menu screen pfod message
  // send menu background, format, prompt, refresh and version
  // parser.print(F("<bg n><b><+6>~"));
  parser.print("`20000"); // refresh each 20sec.
  parser.sendVersion(); // send the menu version
  // send menu items
  parser.print(F("|+A~z"));
#ifdef LUX_GAUGE
  parser.print(F("|+B~lx")); // outside lux
#endif
  parser.print(F("}"));  // close pfod message
}

void sendMainMenuUpdate() {
  parser.print(F("{;"));  // start an Update Menu pfod message
  parser.print(F("~")); // no change to colours and size
  // send menu items
  parser.print(F("|+A~z"));
#ifdef LUX_GAUGE
  parser.print(F("|+B~lx"));
#endif
  parser.print(F("}"));  // close pfod message
  // ============ end of menu ===========
}

static pfodAutoIdx currentTimeIdx;
static pfodAutoIdx dayDateIdx;
static pfodAutoIdx monthYearIdx;
// this assumes pushZero to center of first line
void sendDayDate() {
  dwgs.label().idx(currentTimeIdx).text(getCurrentTime_hhmmAM_PM().c_str()).fontSize(6).bold().color(dwgs.RED).send();
  cSF(sfLabel, 20);
  sfLabel = getCurrentDayOfWeek().c_str();
  sfLabel += ", ";
  sfLabel += getCurrentDayOfMonth().c_str();
  dwgs.label().idx(dayDateIdx).offset(0, 4).text(sfLabel.c_str()).fontSize(1).bold().color(dwgs.RED).send();
  sfLabel = getCurrentMonth().c_str();
  sfLabel += " ";
  sfLabel += getCurrentYear().c_str();
  dwgs.label().idx(monthYearIdx).offset(0, 7).text(sfLabel.c_str()).fontSize(1).bold().color(dwgs.RED).send();
}

void sendDrawing_lx(long dwgIdx) {
  if (dwgIdx == 0) {
    sendDrawing_lx0();
  } else {
    parser.print(F("{}")); // always return somthing
  }
}

void sendDrawing_lx0() {
  dwgs.start(50, 12, dwgs.WHITE, false); //false no more to come,  background defaults to WHITE if omitted i.e. dwgs.start(50,30);
  parser.sendRefreshAndVersion(0); // send the parser version to cache this image, do not auto-refresh main menu refreshes every 20sec
  dwgs.pushZero(6, 5, 0.15); // scale by 1.0f,  scale defaults to 1.0f if omitted, i.e. pushZero(10,15);
  out_lx.draw();  //  gauge
  dwgs.popZero();
  dwgs.end();
}

void sendDrawingUpdates_lx() {
  dwgs.startUpdate();  // dwg refresh updates must be <1024 bytes, i.e. can only be 1 msg
  out_lx.update();
  dwgs.end();
}

void sendDrawing_y(long dwgIdx) {
  if (dwgIdx == 0) {
    sendDrawing_y0();
  } else {
    parser.print(F("{}")); // always return somthing
  }
}

void sendDrawing_y0() {
  dwgs.start(50, 45, dwgs.WHITE, false); //true means more to come,  background defaults to WHITE if omitted i.e. dwgs.start(50,30);
  parser.sendRefreshAndVersion(0); // send the parser version to cache this image, do not auto-refresh main menu refreshes every 20sec
  dwgs.pushZero(40, 2, 0.1); // scale defaults to 1.0f if omitted, i.e. pushZero(10,15);
  dwgs.insertDwg().loadCmd('x').offset(0, 0).send();
  dwgs.popZero(); // back to absolute coordinates and scaling of 1
  dwgs.end();
}


void sendDrawingUpdates_y() {
  dwgs.startUpdate();  // dwg refresh updates must be <1024 bytes, i.e. can only be 1 msg
  dwgs.insertDwg().loadCmd('x').offset(0, 0).send();
  dwgs.end();
}


void sendDrawing_x(long dwgIdx) {
  if (dwgIdx == 0) {
    sendDrawing_x0();
  } else if (dwgIdx == 1) {
    sendDrawing_x1();
  } else {
    parser.print(F("{}")); // always return somthing
  }
}
void sendDrawing_x0() {
  dwgs.start(120, 80, dwgs.WHITE, true); //true means more to come,  background defaults to WHITE if omitted i.e. dwgs.start(50,30);
  parser.sendRefreshAndVersion(0); // send the parser version to cache this image, do not auto-refresh main menu refreshes every 5sec
  dwgs.pushZero(8, 5); // scale defaults to 1.0f if omitted, i.e. pushZero(10,15);
  hPa_graph.draw();  //  RH inside gauge
  dwgs.popZero(); // back to absolute coordinates and scaling of 1
  dwgs.end();
}

void sendDrawing_x1() {
  dwgs.startUpdate(true); //true means more to come,
  dwgs.pushZero(8, 5); // scale defaults to 1.0f if omitted, i.e. pushZero(10,15);
  hPa_graph.update();  //  RH inside gauge
  dwgs.popZero(); // back to absolute coordinates and scaling of 1
  dwgs.end();
}


void sendDrawingUpdates_x() {
  dwgs.startUpdate();  // dwg refresh updates must be <1024 bytes, i.e. can only be 1 msg
  hPa_graph.update(); // see graph for update code
  dwgs.end();
}

// msgs are limited to 1023 bytes, so dwgs are often sent in blocks
void sendDrawing_z(long dwgIdx) {
  if (dwgIdx == 0) {
    sendDrawing_z0();
  } else if (dwgIdx == 1) {
    sendDrawing_z1();
  } else if (dwgIdx == 2) {
    sendDrawing_z2();
  } else if (dwgIdx == 3) {
    sendDrawing_z3();
  } else if (dwgIdx == 4) {
    sendDrawing_z4();
  } else {
    parser.print(F("{}")); // always return somthing
  }
}

void sendDrawing_z0() {
  dwgs.start(50, 65, dwgs.WHITE, true); //true means more to come,  background defaults to WHITE if omitted i.e. dwgs.start(50,30);
  parser.sendRefreshAndVersion(0); // send the parser version to cache this image, do not auto-refresh main menu refreshes every 20sec
  dwgs.pushZero(12.5, 43, 1.35f); // scale defaults to 1.0f if omitted, i.e. pushZero(10,15);
  in_rh.draw();  //  RH inside gauge
  dwgs.popZero(); // back to absolute coordinates and scaling of 1
  dwgs.end();
}

void sendDrawing_z1() {
  dwgs.startUpdate(true); //true means more to come,
  dwgs.pushZero(37.5, 43, 1.35f); // scale defaults to 1.0f if omitted, i.e. pushZero(10,15);
  out_rh.draw();  //  Rh outside gauge
  dwgs.popZero(); // back to absolute coordinates and scaling of 1
  dwgs.end();
}

void sendDrawing_z2() {
  dwgs.startUpdate(true); //true means more to come,
  dwgs.label().offset(12.5, 2.5).text("Inside").fontSize(2).bold().color(dwgs.BLACK).send();
  cSF(sfText, 30);
  sfText = "W_1";
  getLastSeenMins(sfText, lastSeen_in);  // if missed sensor reading add time since last one received.
  dwgs.label().idx(z_idx + 1).fontSize(-6).bold().text(sfText.c_str()).color(dwgs.BLACK).offset(20, 33).send(); // add index so this is updated
  dwgs.pushZero(20, 31); // scale by 1.0f,  scale defaults to 1.0f if omitted, i.e. pushZero(10,15);
  inside_temp.draw();  // draw inside temp gauge
  dwgs.popZero(); // back to absolute coordinates and scaling of 1
  dwgs.pushZero(25, 31); // scale by 1.0f,  scale defaults to 1.0f if omitted, i.e. pushZero(10,15);
  inside_temp.drawScale(); // draw scale numbers in center.
  dwgs.popZero();
  dwgs.end();
}

void sendDrawing_z3() {
  dwgs.startUpdate(true); //true means more to come,
  dwgs.label().offset(37.5, 2.5).text("Outside").fontSize(2).bold().color(dwgs.BLACK).send();
  cSF(sfText, 30);
  sfText = "WL_1";
  getLastSeenMins(sfText, lastSeen_out);
  dwgs.label().idx(z_idx + 2).fontSize(-6).bold().text(sfText.c_str()).color(dwgs.BLACK).offset(30, 33).send(); // add index so this is updated
  dwgs.pushZero(30, 31); // scale by 1.0f,  scale defaults to 1.0f if omitted, i.e. pushZero(10,15);
  outside_temp.draw();  // draw outside gauge
  dwgs.popZero();
  dwgs.end();
}

void sendDrawing_z4() {
  dwgs.startUpdate(false); //true means more to come,
  // add plot touch zone over the whole dwg. So any click anywhere on the dwg will open the plots
  dwgs.pushZero(25,49.5,0.4);
  chartButton.draw(); // add chartButton touchZone AFTER global one so it is on top
  dwgs.popZero();
  dwgs.pushZero(59, 127, 0.07); // scale defaults to 1.0f if omitted, i.e. pushZero(10,15);
  dwgs.insertDwg().loadCmd('x').offset(0, 0).send();
  dwgs.popZero(); // back to absolute coordinates and scaling of 1
  dwgs.pushZero(12.5,56);
  sendDayDate();
  dwgs.popZero();
  dwgs.end();
}

// update existing dwg.  uses z_idx to match update to existing dwg item
void sendDrawingUpdates_z() {
  dwgs.startUpdate();  // dwg refresh updates must be <1024 bytes, i.e. can only be 1 msg
  cSF(sfText, 30);
  sfText = "W_1";
  getLastSeenMins(sfText, lastSeen_in);  // if missed sensor reading add time since last one received.
  dwgs.label().idx(z_idx + 1).fontSize(-6).bold().text(sfText.c_str()).color(dwgs.BLACK).offset(20, 33).send(); // add index so this is updated
  sfText = "WL_1";
  getLastSeenMins(sfText, lastSeen_out);
  dwgs.label().idx(z_idx + 2).fontSize(-6).bold().text(sfText.c_str()).color(dwgs.BLACK).offset(30, 33).send(); // add index so this is updated
  in_rh.update(); // see gauge for update code
  out_rh.update();
  outside_temp.update();
  inside_temp.update();
  sendDayDate();
  dwgs.insertDwg().loadCmd('x').offset(0, 0).send();
  dwgs.end();
}


int swap01(int in) {
  return (in == 0) ? 1 : 0;
}
// ============= end generated code =========
