/*
   hPa_BarGraph.cpp
   (c)2014-2022 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.
*/

#include "hPa_BarGraph.h"
#include "SafeString.h"
#include "millisDelay.h"

// barometer guage from 960 to 1070 hPa
// value (int) ranges from 9600 to 10700
//  static const unsigned int minValueTimes10 = 9900;
//  static const unsigned int maxValueTimes10 = 10250;

const float hPa_BarGraph::fullScaleLength = 100.0;
static float hrScale = 2.33 * 48.0/hPa_BarGraph::GRAPH_X_AXIS_SIZE;
static float hPaScale = 0.5;

hPa_BarGraph::hPa_BarGraph(pfodDwgs *_dwgsPtr) : pfodControl(_dwgsPtr) {
  z_idx = pfodDwgs::reserveIdx(getArraySize() + 10); // z_idx, to z_idx+getArraySize()-1 to getArraySize()+9
  for (size_t i = 0; i < getArraySize(); i++) {
    hPa_times10_Array[i] = 0; // invalid
    hPa_time_Array[i] = 0; // invalid
  }
  set_hPa(0.0);
}

size_t hPa_BarGraph::getArraySize() {
  return GRAPH_X_AXIS_SIZE; // 0 to ARRAY_SIZE-1
}

long hPa_BarGraph::getDataInterval_ms() {
  return X_AXIS_MS / getArraySize();
}

// this called to fill entire data array
void hPa_BarGraph::setArrayData(unsigned long ms, unsigned int hPa_Times10, size_t idx) {
  if (idx >= getArraySize()) {
    return; // outof bounds
  }
  hPa_time_Array[idx] = ms;
  hPa_times10_Array[idx] = hPa_Times10;
}

void hPa_BarGraph::set_hPa(float hPaValue) {
  setValue(hPaValue * 10.0);
}

// 10*hPa == value
float hPa_BarGraph::valueToLength(unsigned int valueTimes10) {
  if (valueTimes10 < minValueTimes10) {
    return 0;
  }
  if (valueTimes10 > maxValueTimes10) {
    return fullScaleLength * hPaScale;
  }
  // else
  float len = (float)(valueTimes10 - minValueTimes10) / (float)(maxValueTimes10 - minValueTimes10) * fullScaleLength;
  if (len > fullScaleLength) {
    len = fullScaleLength;
  }
  if (len < 0) {
    len = 0;
  }
  return len * hPaScale;
}

// scale to x Axis -24hrs to 0
// deltaT_ms is +ve from 0 to 24hrs in ms
// time to offset is t_ms/XAXIS_MS * hrScale *getArraySize();
float hPa_BarGraph::timeToXaxis(unsigned long deltaT_ms) {
  deltaT_ms = deltaT_ms;
  float rtn = 0;
  if (deltaT_ms >= X_AXIS_MS) {
    return rtn;
  }
  if (deltaT_ms == 0) {
    rtn = hrScale * getArraySize();
    return rtn;
  }
  // else
  rtn = ((float)( X_AXIS_MS - deltaT_ms)) / ((float)X_AXIS_MS) * hrScale * getArraySize();
  return rtn;
}

void hPa_BarGraph::drawTimeScalePoint(int hrs) {
  // zero at the bottom
  if ((hrs < - 24) || (hrs > 0)) {
    return;
  }
  float xOffset = ((float)(hrs + 24)) / 24.0 * hrScale * (getArraySize());
  cSF(sfLabel, 20);
  sfLabel = hrs;
  if ((hrs == -24) || (hrs == -12) || (hrs == 0)) {
    sfLabel += "hrs";
    dwgsPtr->label().offset(xOffset, valueToLength(maxValueTimes10) + (hPaScale * 15)).bold().fontSize(8).text(sfLabel.c_str()).send();
  } else {
    dwgsPtr->label().offset(xOffset, valueToLength(maxValueTimes10) + (hPaScale * 10)).bold().fontSize(8).text(sfLabel.c_str()).send();
  }
  if (hrs != -24) {
    dwgsPtr->line().color(dwgsPtr->SILVER).offset(xOffset, valueToLength(maxValueTimes10)).size(0, -valueToLength(maxValueTimes10)).send();
  }
}
void hPa_BarGraph::drawScalePoint(int hPa) {
  unsigned int hPaTimes10 = hPa * 10;
  if (hPaTimes10 < minValueTimes10) {
    return;
  }
  if (hPaTimes10 > maxValueTimes10) {
    return;
  }
  cSF(sfLabel, 20);
  sfLabel = hPa;
  //sfLabel += "hPa";
  float yOffset = valueToLength(maxValueTimes10) - valueToLength(hPaTimes10);
  float xOffset = hrScale * getArraySize();
  if (hPa % 10 == 0) {
    dwgsPtr->label().offset(xOffset + 11.65, yOffset).bold().fontSize(8).text(sfLabel.c_str()).send();
  } else {
    dwgsPtr->label().offset(xOffset + 11.65, yOffset).bold().fontSize(6).text(sfLabel.c_str()).send();
  }
  if (hPaTimes10 <= minValueTimes10) {
    return;
  }
  if (hPaTimes10 >= maxValueTimes10) {
    return;
  }
  dwgsPtr->line().color(dwgsPtr->SILVER).offset(hrScale * getArraySize(), yOffset).size(-(hrScale * getArraySize()), 0).send();
}

void hPa_BarGraph::drawScale() {
  // labels/lines at each 10hPa
  for (int i = 960; i <= 1070; i = i + 10) {
    drawScalePoint(i);
  }
  for (int i = -24; i <= 0; i = i + 4) {
    drawTimeScalePoint(i);
  }
  dwgsPtr->label().offset(hrScale * getArraySize() + 11.65, -6).bold().fontSize(8).text("hPa").send();
}

void hPa_BarGraph::draw() {
  drawScale();
}

// call this in its own sendDrawing_x.. as it is it almost 1024 long
void hPa_BarGraph::update() {
  unsigned long ms = millis();
  for (size_t i = 0; i < getArraySize(); i++) {
    if ((hPa_times10_Array[i] < minValueTimes10) || (hPa_times10_Array[i] > maxValueTimes10)) {
      dwgsPtr->index().idx(z_idx + i).send(); // reserve idx
      continue; // skip this one
    }

    float nextXPosition = timeToXaxis(ms - hPa_time_Array[i]);
    float nextYPosition = valueToLength(maxValueTimes10) - valueToLength(hPa_times10_Array[i]);
    dwgsPtr->circle().filled().offset(nextXPosition, nextYPosition).idx(z_idx + i).send(); //.radius(hrScale*0.25)
  }

  // final point is current value
  float nextXPosition = timeToXaxis(0);
  float nextYPosition = valueToLength(maxValueTimes10) - valueToLength(getValue());
  if ((getValue() < minValueTimes10) || (getValue() > maxValueTimes10)) {
  	// skip this point
  } else {
    dwgsPtr->circle().filled().offset(nextXPosition, nextYPosition).idx(z_idx + getArraySize()).send(); //.radius(hrScale*0.25)
  }
  { // Always show the current value at the extreme range of the graph even if point is off scale
    cSF(sfLabel, 20);
    sfLabel.print(((float)getValue()) / 10.0, 1);
    dwgsPtr->rectangle().filled().idx(z_idx + getArraySize() + 1).color(dwgsPtr->WHITE).offset(hrScale * getArraySize() + 2.33, nextYPosition - 4).size(30, 8).send();
    dwgsPtr->label().idx(z_idx + getArraySize() + 2).color(dwgsPtr->RED).left().offset(hrScale * getArraySize() + 2.91, nextYPosition).bold().fontSize(10).text(sfLabel.c_str()).send();
  }
}
