/*
   hPa_Gauge.cpp
   (c)2014-2022 Forward Computing and Control Pty. Ltd.
   NSW Australia, www.forward.com.au
   This code is not warranted to be fit for any purpose. You may only use it at your own risk.
   This code may be freely used for both private and commercial use
   Provide this copyright is maintained.
*/

#include "hPa_Gauge.h"
#include "SafeString.h"
// barometer guage from 960 to 1070 hPa
// value (int) ranges from 9600 to 10700

hPa_Gauge::hPa_Gauge(pfodDwgs *_dwgsPtr) : pfodControl(_dwgsPtr) {
  z_idx = pfodDwgs::reserveIdx(6); // z_idx, to z_idx+5
  value_previous = 0; // invalid
  cmd = 0; // invalid
}

void hPa_Gauge::set_hPa(float hPaValue) {
  setValue(hPaValue * 10);
}

void hPa_Gauge::setPrevious_hPa(float hPaValue) {
  value_previous = hPaValue * 10;
  if (value_previous > 0) {
    markerSet_ms = millis();
  }
}

// 10*hPa == value == 9600 to 10600 so angle = (int)((float)(value)-9600.0 * 320.0)/1000.0
static float valueToAngle(float value) {
  float ang = (((float)value) - 9600.0) * 320.0 / 1000.0;
  if (ang > 320) {
    ang = 320;
  }
  if (ang < 0) {
    ang = 0;
  }
  return ang;
}

void hPa_Gauge::setCmd(char _cmd) {
  cmd = _cmd;
}

void hPa_Gauge::drawScalePoint(int hPa) {
  cSF(sfLabel, 20);
  sfLabel = hPa;
  float ang = 250 - valueToAngle(hPa * 10);
  dwgsPtr->arc().filled().color(pfodDwgs::BLACK).radius(5.5).start(ang).angle(0).send();
  if (hPa < 1010) {
    dwgsPtr->label().offset(dwgsPtr->xRadius(ang, 6), dwgsPtr->yRadius(ang, 6)).right().fontSize(-6).text(sfLabel.c_str()).send();
  } else if (hPa > 1010) {
    dwgsPtr->label().offset(dwgsPtr->xRadius(ang, 6), dwgsPtr->yRadius(ang, 6)).left().fontSize(-6).text(sfLabel.c_str()).send();
  } else { // == 1010
    dwgsPtr->label().offset(dwgsPtr->xRadius(ang, 6), dwgsPtr->yRadius(ang, 6) - 0.15).fontSize(-6).text(sfLabel.c_str()).send();
  }
}

void hPa_Gauge::draw() {
  dwgsPtr->arc().filled().color(dwgsPtr->GREY).radius(5).start(250).angle(-320).send();
  dwgsPtr->arc().filled().color(dwgsPtr->WHITE).radius(3).idx(z_idx + 1).start(250).angle(-320).send();

  // labels at 970, 1010, 1050
  // draw lines
  drawScalePoint(970);
  drawScalePoint(980);
  drawScalePoint(990);
  drawScalePoint(1000);
  drawScalePoint(1010);
  drawScalePoint(1020);
  drawScalePoint(1030);
  drawScalePoint(1040);
  drawScalePoint(1050);
  if (cmd != 0) { // have set a cmd
    // add touch zone for setting pointer
    // =============== start active area
    // define button active region and buttonCmd
    dwgsPtr->touchZone().cmd(cmd).size(12, 12).centered().filter(dwgsPtr->PRESS).send();
    // ========== end active area
  }

  update(); // update with current state
}

void hPa_Gauge::update() {
  // angle 0 to 340 => value 9600 to 10600 so angle = (int)((float)(value)-9600.0 * 340.0)/1000.0
  dwgsPtr->arc().filled().color(dwgsPtr->RED).radius(5).idx(z_idx).start(250).angle(-valueToAngle(value)).send();
  dwgsPtr->label().color(dwgsPtr->RED).offset(0, 7.5).idx(z_idx + 2).fontSize(1).bold().floatReading(((float)value) / 10.0).decimals(1).units(" hPa").send();
  //float angValue = 250 - valueToAngle(value);
  // dwgsPtr->line().color(dwgsPtr->BLACK).idx(z_idx + 3).size(dwgsPtr->xRadius(angValue, 5.5), dwgsPtr->yRadius(angValue, 5.5)).send();
  unsigned long hrs = 0;
  int mins = 0;
  if (value_previous > 0) {
    // show marked hPa and time since
    unsigned long mins_sinceMarked = (millis() - markerSet_ms) / (60ul * 1000);
    hrs = mins_sinceMarked / 60;
    if (hrs >= 99) {
      // reset mark
      value_previous = 0;
    } else {
      mins = mins_sinceMarked - hrs * 60;
    }
  }
  if (value_previous > 0) {
    float angPrevious = 250 - valueToAngle(value_previous);
    dwgsPtr->line().color(dwgsPtr->BLACK).idx(z_idx + 4).size(dwgsPtr->xRadius(angPrevious, 2.85), dwgsPtr->yRadius(angPrevious, 2.85)).send();
    cSF(sfText, 28);
    sfText.print(((float)value_previous) / 10.0, 1); // 4+dot+1 = 6
    sfText += " hPa  "; // 6 + 6 = 12
    sfText += hrs; // 14
    sfText += ":"; // 15
    if (mins < 10) {
      sfText += '0';
    }
    sfText += mins; // 17
    sfText += " hrs ago"; // 24
    dwgsPtr->label().idx(z_idx + 5).fontSize(-7).bold().text(sfText.c_str()).color(dwgsPtr->BLACK).offset(0, 10).send();
  } else {
    float angValue = 250 - valueToAngle(value);
    dwgsPtr->line().color(dwgsPtr->BLACK).idx(z_idx + 4).size(dwgsPtr->xRadius(angValue, 2.85), dwgsPtr->yRadius(angValue, 2.85)).send();
    dwgsPtr->label().idx(z_idx + 5).fontSize(-7).bold().text("Press Gauge to freeze Marker").color(dwgsPtr->BLACK).offset(0, 10).send();
  }

}
